
/*****************************************************************************

This class has been written by Elm (elmue@gmx.de)

Check if you have the latest version on:
https://www.codeproject.com/Articles/5293980/Graph3D-A-Windows-Forms-Render-Control-in-Csharp

*****************************************************************************/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

using delRendererFunction = Plot3D.Graph3D.delRendererFunction;
using eRaster             = Plot3D.Graph3D.eRaster;
using cPoint3D            = Plot3D.Graph3D.cPoint3D;
using cMessgData          = Plot3D.Graph3D.cMessgData;
using cSurfaceData        = Plot3D.Graph3D.cSurfaceData;
using cScatterData        = Plot3D.Graph3D.cScatterData;
using cLineData           = Plot3D.Graph3D.cLineData;
using cPolygonData        = Plot3D.Graph3D.cPolygonData;
using cColorScheme        = Plot3D.Graph3D.cColorScheme;
using eNormalize          = Plot3D.Graph3D.eNormalize;
using eSurfaceMode        = Plot3D.Graph3D.eSurfaceMode;
using eScatterMode        = Plot3D.Graph3D.eScatterMode;
using eScatterShape       = Plot3D.Graph3D.eScatterShape;
using eSchema             = Plot3D.ColorSchema.eSchema;

namespace Plot3D
{
    public partial class Graph3DMainForm : Form
    {
        public Graph3DMainForm()
        {
            InitializeComponent();
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            // This is optional. If you don't want to use Trackbars leave this away.
            graph3D.AssignTrackBars(trackRho, trackTheta, trackPhi);

            graph3D.BeginUpdate();
            cMessgData i_Mesg1 = new cMessgData("Please select a Data Source in the combobox at the left", 10, 10, Color.Blue);
            cMessgData i_Mesg2 = new cMessgData("(This is a user message that you can change)",            10, 27, Color.LightGray);
            graph3D.AddMessageData(i_Mesg1);
            graph3D.AddMessageData(i_Mesg2);
            graph3D.EndUpdate(eNormalize.Separate);

            comboRaster.Sorted = false;
            foreach (eRaster e_Raster in Enum.GetValues(typeof(eRaster)))
            {
                comboRaster.Items.Add(e_Raster);
            }
            comboRaster.SelectedIndex = (int)eRaster.Labels;

            comboColors.Sorted = false;
            foreach (eSchema e_Schema in Enum.GetValues(typeof(eSchema)))
            {
                comboColors.Items.Add(e_Schema);
            }
            comboColors.SelectedIndex = (int)eSchema.Rainbow1;

            comboDataSrc.SelectedIndex = -1; // set empty
        }

        private void comboColors_SelectedIndexChanged(object sender, EventArgs e)
        {
            DrawGraph();
        }
        private void comboDataSrc_SelectedIndexChanged(object sender, EventArgs e)
        {
            DrawGraph();
        }

        private void DrawGraph()
        {
            switch (comboDataSrc.SelectedIndex)
            {
                case  0: LoadCallback();                        break; // Callback
                case  1: LoadFormula();                         break; // Formula
                case  2: LoadSurface(eSurfaceMode.Fill);        break; // Surface Fill
                case  3: LoadSurface(eSurfaceMode.Lines);       break; // Surface Grid
                case  4: LoadNestedGraphs();                    break; // Nested Graphs
                case  5: LoadScatterPlot(false);                break; // Scatter Plot
                case  6: LoadScatterPlot(true);                 break; // Scatter Lines
                case  7: LoadScatterShapes();                   break; // Scatter Shapes
                case  8: LoadPyramid();                         break; // Pyramid
                case  9: LoadSphere(eSurfaceMode.Fill,  true);  break; // Sphere Fill Closed
                case 10: LoadSphere(eSurfaceMode.Fill,  false); break; // Sphere Fill Open
                case 11: LoadSphere(eSurfaceMode.Lines, true);  break; // Sphere Grid
                case 12: LoadValentine();                       break; // Valentine's heart
            }

            lblInfo.Text = graph3D.ObjectCount;
        }

        private void comboRaster_SelectedIndexChanged(object sender, EventArgs e)
        {
            graph3D.Raster = (eRaster)comboRaster.SelectedIndex;
        }

        private void btnReset_Click(object sender, EventArgs e)
        {
            graph3D.SetCoefficients(1350, 70, 230);
        }

        private void btnScreenshot_Click(object sender, EventArgs e)
        {
            SaveFileDialog i_Dlg = new SaveFileDialog();
            i_Dlg.Title      = "Save as PNG image";
            i_Dlg.Filter     = "PNG Image|*.png";
            i_Dlg.DefaultExt = ".png";

            if (DialogResult.Cancel == i_Dlg.ShowDialog(this))
                return;

            Bitmap i_Bitmap = graph3D.GetScreenshot();
            try
            {
                i_Bitmap.Save(i_Dlg.FileName, ImageFormat.Png);
            }
            catch (Exception Ex)
            {
                MessageBox.Show(this, Ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // ================================================================================================

        /// <summary>
        /// This demonstrates how to use a callback function which calculates Z values from X and Y
        /// </summary>
        private void LoadCallback()
        {
            Color[]      c_Colors = ColorSchema.GetSchema((eSchema)comboColors.SelectedIndex);
            cColorScheme i_Scheme = new cColorScheme(1, c_Colors);

            delRendererFunction f_Callback = delegate(double X, double Y)
            {
                double r = 0.15 * Math.Sqrt(X * X + Y * Y);
                if (r < 1e-10) return 120;
                else           return 120 * Math.Sin(r) / r;
            };

            cMessgData i_Mesg1 = new cMessgData("r = 0.15 * sqrt(x * x + y * y)", 10, -27, Color.Indigo);
            cMessgData i_Mesg2 = new cMessgData("z = 120  * sin(r) / r",          10, -10, Color.Indigo);

            // IMPORTANT: Normalize maintainig the relation between X,Y,Z values otherwise the function will be distorted.
            graph3D.BeginUpdate();
            graph3D.AddMessageData(i_Mesg1);
            graph3D.AddMessageData(i_Mesg2);
            graph3D.AddFunctionData(f_Callback, new PointF(-120, -80), new PointF(120, 80), 5, eSurfaceMode.Fill, i_Scheme);
            graph3D.EndUpdate(eNormalize.MaintainXYZ);
        }

        /// <summary>
        /// This demonstrates how to use a string formula which calculates Z values from X and Y
        /// </summary>
        private void LoadFormula()
        {
            Color[]      c_Colors = ColorSchema.GetSchema((eSchema)comboColors.SelectedIndex);
            cColorScheme i_Scheme = new cColorScheme(1, c_Colors);

            String  s_Formula = "12 * sin(x) * cos(y) / (sqrt(sqrt(x * x + y * y)) + 0.2)";
            cMessgData i_Mesg = new cMessgData("Formula: z = " + s_Formula, 10, -10, Color.Indigo);

            try
            {
                delRendererFunction f_Function = FunctionCompiler.Compile(s_Formula);

                // IMPORTANT: Normalize maintainig the relation between X,Y,Z values otherwise the function will be distorted.
                graph3D.BeginUpdate();
                graph3D.AddMessageData(i_Mesg);
                graph3D.AddFunctionData(f_Function, new PointF(-10, -10), new PointF(10, 10), 0.5, eSurfaceMode.Fill, i_Scheme);
                graph3D.EndUpdate(eNormalize.MaintainXYZ);
            }
            catch (Exception Ex) // invalid formula
            {
                MessageBox.Show(Ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// This demonstrates how to set X, Y, Z values directly (without function)
        /// </summary>
        private void LoadSurface(eSurfaceMode e_Surface)
        {
            #region int s32_Values definition

            int[,] s32_Values = new int[,]
            {
                { 34767, 34210, 33718, 33096, 32342, 31851, 31228, 30867, 31457, 30867, 30266, 28934, 27984, 26492, 25167, 25167, 25167},
                { 34669, 34210, 33653, 33096, 32539, 32047, 31490, 30933, 31293, 30671, 29983, 28803, 27886, 26492, 25167, 25167, 25167},
                { 34603, 34144, 33718, 33227, 32768, 32342, 31719, 30999, 31228, 30333, 29622, 28606, 27886, 26492, 25167, 25167, 25167},
                { 34472, 34079, 33653, 33161, 32768, 32408, 31785, 31162, 30802, 30048, 29360, 28312, 27755, 26367, 25049, 25049, 25049},
                { 34210, 33784, 33423, 33161, 32801, 32408, 31785, 31097, 30474, 29622, 29000, 28115, 27623, 26367, 25049, 25049, 25049},
                { 33980, 33587, 33161, 32935, 32588, 32342, 31621, 30802, 29852, 29000, 28377, 27689, 27421, 26367, 25049, 25049, 25049},
                { 33522, 33227, 32702, 32615, 32452, 31851, 30933, 30179, 29295, 28358, 27984, 27132, 27301, 26367, 25049, 25049, 25049},
                { 32672, 32178, 31916, 31469, 31246, 30540, 29852, 29065, 28377, 27623, 27263, 26706, 26935, 26367, 25049, 25049, 25049},
                { 30769, 30423, 29917, 29231, 29392, 28705, 28075, 27726, 27263, 26691, 26417, 26182, 26575, 26575, 25246, 25246, 25246},
                { 27525, 27518, 26701, 27334, 27682, 27402, 26903, 26707, 26444, 25887, 25719, 25690, 26122, 26122, 26122, 26122, 26122},
                { 23475, 23888, 24478, 25330, 26212, 26199, 25701, 25664, 25740, 25013, 24904, 25068, 25374, 25374, 25374, 25374, 25374},
                { 20677, 21445, 22544, 23593, 24441, 24785, 24538, 24644, 24773, 24299, 24062, 24576, 24510, 24510, 24510, 24510, 24510},
                { 18743, 19792, 20808, 22086, 22805, 23167, 23486, 23366, 23757, 23411, 23691, 23822, 23822, 23822, 23822, 23822, 23822},
                { 17334, 18579, 19497, 20775, 21463, 21848, 22288, 22446, 22643, 22446, 22643, 22708, 23069, 23069, 23069, 23069, 23069},
                { 16155, 17236, 18350, 19399, 20251, 20677, 21016, 21332, 21660, 21791, 21889, 21955, 22217, 22217, 22217, 22217, 22217},
                { 14746, 15860, 17039, 17990, 18842, 19595, 20050, 20349, 20546, 20840, 20972, 20972, 21332, 21332, 21332, 21332, 21332},
                { 13337, 14516, 15729, 16679, 17564, 18514, 18907, 19169, 19399, 19661, 19792, 19594, 20152, 20152, 20152, 20152, 20152},
                { 12452, 13435, 14615, 15499, 16253, 17105, 17596, 17924, 18153, 18285, 18428, 18776, 19104, 19104, 19104, 19104, 19104},
                { 11469, 12354, 13533, 14287, 15008, 15925, 16187, 16482, 16690, 16976, 17105, 17302, 17531, 17531, 17531, 17531, 17531},
                { 10486, 11370, 12255, 13009, 13861, 14746, 15172, 15368, 15434, 15630, 15794, 15991, 16351, 16351, 16351, 16351, 16351},
                {  9684, 10387, 11141, 11796, 12546, 13337, 14029, 14320, 14549, 14811, 14939, 15434, 15794, 15794, 15794, 15794, 15794},
                {  9059,  9634, 10617, 11141, 11838, 12681, 13411, 13861, 14121, 14624, 14868, 15172, 15368, 15368, 15368, 15368, 15368},
            };

            #endregion

            Color[]      c_Colors = ColorSchema.GetSchema((eSchema)comboColors.SelectedIndex);
            cColorScheme i_Scheme = new cColorScheme(2, c_Colors);
            cSurfaceData i_Data   = new cSurfaceData(e_Surface, s32_Values.GetLength(0), s32_Values.GetLength(1), i_Scheme);

            for (int C=0; C<i_Data.Cols; C++)
            {
                for (int R=0; R<i_Data.Rows; R++)
                {
                    double d_X = C *  10.0;
                    double d_Y = R * 500.0;
                    double d_Z = s32_Values[C,R] / 327.68;

                    i_Data.SetPointAt(C, R,  d_X, d_Y, d_Z);
                }
            }

            // IMPORTANT: Normalize X,Y,Z separately because the axes have different ranges
            graph3D.BeginUpdate();
            graph3D.SetAxisLegends("MAP (kPa)", "Engine Speed (rpm)", "Volume Efficiency (%)");
            graph3D.AddRenderData(i_Data);
            graph3D.EndUpdate(eNormalize.Separate);
        }

        /// <summary>
        /// This loads 2 graphs, one nested into the other
        /// </summary>
        private void LoadNestedGraphs()
        {
            const int POINTS = 8;
            cSurfaceData i_Data1 = new cSurfaceData(eSurfaceMode.Lines, POINTS, POINTS, new cColorScheme(3, Color.Orange));
            cSurfaceData i_Data2 = new cSurfaceData(eSurfaceMode.Lines, POINTS, POINTS, new cColorScheme(2, Color.Black));

            for (int C=0; C<POINTS; C++)
            {
                for (int R=0; R<POINTS; R++)
                {
                    double d_X = (C - POINTS / 2.3) / (POINTS / 5.5);
                    double d_Y = (R - POINTS / 2.3) / (POINTS / 5.5);
                    double d_Radius = Math.Sqrt(d_X * d_X + d_Y * d_Y);
                    double d_Z = Math.Cos(d_Radius) + 1.0;

                    i_Data1.SetPointAt(C, R,  d_X, d_Y, d_Z);
                    i_Data2.SetPointAt(C, R,  d_X, d_Y, d_Z * 0.6);
                }
            }

            cMessgData i_Mesg1 = new cMessgData("Graph with error data",   10, -10, Color.Orange);
            cMessgData i_Mesg2 = new cMessgData("Graph with correct data", 10, -27, Color.Black);

            graph3D.BeginUpdate();
            graph3D.AddRenderData (i_Data1);
            graph3D.AddRenderData (i_Data2);
            graph3D.AddMessageData(i_Mesg1);
            graph3D.AddMessageData(i_Mesg2);
            graph3D.EndUpdate(eNormalize.MaintainXY);
        }

        /// <summary>
        /// This demonstrates how to set X, Y, Z scatterplot points in form of a spiral.
        /// </summary>
        private void LoadScatterPlot(bool b_Lines)
        {
            // 3 pixels for line width and for circle radius
            const int SIZE = 3;

            Color[]    c_Colors = ColorSchema.GetSchema((eSchema)comboColors.SelectedIndex);
            eScatterMode e_Mode = b_Lines ? eScatterMode.Lines : eScatterMode.Shapes;
            cScatterData i_Data = new cScatterData(e_Mode, new cColorScheme(SIZE, c_Colors));

            for (double P = -22.0; P < 22.0; P += 0.1)
            {
                double d_X = Math.Sin(P) * P;
                double d_Y = Math.Cos(P) * P;
                double d_Z = P;
                if (d_Z > 0.0) d_Z/= 3.0;

                if (b_Lines) i_Data.AddLine (d_X, d_Y, d_Z, true, null);
                else         i_Data.AddShape(d_X, d_Y, d_Z, eScatterShape.Circle, SIZE, null);
            }

            cMessgData i_Mesg = new cMessgData("Move the mouse over a scatter point to see the tooltip", 10, -10, Color.Gray);

            // Depending on your use case you can also specify MaintainXY or MaintainXYZ here
            graph3D.BeginUpdate();
            graph3D.AddRenderData(i_Data);
            graph3D.AddMessageData(i_Mesg);
            graph3D.EndUpdate(eNormalize.Separate);
        }

        private void LoadScatterShapes()
        {
            #region double d_Values definitions

            double[,] d_Values = new double[,]
            {
                // Value  X        Y      Z
                {   0.39, 0.0051,  0.133, 0.66 },
                {   0.23, 0.0002,  0.114, 0.87 },
                {   1.46, 0.0007,  0.077, 0.72 },
                {  -1.85, 0.0137,  0.053, 0.87 },
                {   5.51, 0.0047,  0.016, 1.12 },
                {   1.15, 0.0076,  0.117, 1.36 },
                {   1.98, 0.0157, -0.004, 1.23 },
                {  -2.22, 0.0029,  0.037, 1.09 },
                {   0.94, 0.0066, -0.006, 0.91 },
                {   0.56, 0.0328, -0.115, 1.27 },
                {   4.70, 0.0333, -0.154, 1.38 },
                {  -6.42, 0.0594, -0.228, 2.48 },
                {  -7.93, 0.0487, -0.394, 1.24 },
                {   1.57, 0.0874, -0.504, 0.78 },
                {  -6.92, 0.0739, -0.395, 1.05 },
                {   4.65, 0.0341, -0.484, 2.18 },
                {   7.10, 0.0326, -0.477, 2.00 },
                {   3.31, 0.0024, -0.090, 0.62 },
                {   6.83, 0.0138, -0.045, 1.04 },
                {   0.02, 0.0021,  0.139, 0.60 },
                {   3.71, 0.0137,  0.033, 0.81 },
                {   0.90, 0.0066,  0.120, 1.52 },
                {   0.68, 0.0271,  0.150, 0.74 },
                {   1.95, 0.0043,  0.147, 0.89 },
                {   4.91, 0.0192,  0.046, 1.69 },
                {   0.02, 0.0086,  0.024, 1.32 },
                {  -7.47, 0.0488, -0.021, 1.18 },
                {  -1.09, 0.1051, -0.221, 1.17 },
                {  -1.72, 0.0322, -0.244, 0.95 },
                {   1.83, 0.0078,  0.083, 1.12 },
                {   1.71, 0.0049,  0.080, 0.79 },
                {  -7.24, 0.0012,  0.077, 2.08 },
                {   6.08, 0.0644, -0.131, 1.28 },
                {  -0.38, 0.0000,  0.022, 0.91 },
                {   0.03, 0.0119,  0.146, 0.81 },
                {   1.86, 0.0131,  0.088, 0.69 },
                {   2.80, 0.0010,  0.068, 1.03 },
                {   1.66, 0.0094,  0.158, 1.20 },
                {   1.34, 0.0106,  0.162, 1.06 },
                {   2.36, 0.0090,  0.016, 1.18 },
                {   0.58, 0.0012,  0.128, 1.33 },
                {   4.98, 0.0204,  0.118, 1.36 },
                {   3.02, 0.0314,  0.042, 1.57 },
                {   0.47, 0.0120,  0.180, 0.92 },
                {  -0.41, 0.0008,  0.197, 0.65 },
                {  -7.98, 0.0452, -0.069, 1.06 },
                {   3.45, 0.0900, -0.390, 1.49 },
                {  -6.74, 0.0270, -0.688, 1.64 },
                { -12.86, 0.0538, -0.283, 1.87 },
                {  -9.34, 0.0526, -0.671, 1.56 },
                {  10.03, 0.0389, -0.981, 1.49 },
                {   5.26, 0.0299, -0.463, 1.31 },
                {   8.95, 0.0248, -0.442, 0.78 },
                {   5.51, 0.0182, -0.007, 0.90 },
                {   1.94, 0.0060,  0.356, 0.59 },
                {   1.23, 0.0041,  0.260, 0.97 },
                {  -0.57, 0.0032,  0.468, 0.66 },
                {   0.00, 0.0066,  0.386, 0.72 },
                {  14.45, 0.0526, -0.013, 1.40 },
                {  -7.35, 0.0467,  0.223, 1.45 },
                {  -7.39, 0.0479, -0.138, 0.76 },
                {   2.00, 0.0174, -0.406, 1.05 },
                {   1.70, 0.0159, -0.080, 0.95 },
                {   1.74, 0.0073,  0.060, 0.51 },
                {   0.53, 0.0085, -0.052, 0.74 },
                {   7.04, 0.0567, -0.400, 0.97 },
                {   0.35, 0.0072,  0.139, 0.57 },
                {   1.20, 0.0077,  0.195, 0.98 },
                {   4.47, 0.0043,  0.206, 0.76 },
                {  -0.39, 0.0120,  0.167, 1.24 },
                {   3.85, 0.0297, -0.106, 0.99 },
                {   3.75, 0.0372, -0.085, 1.51 },
                {  -7.03, 0.0149,  0.077, 0.58 },
                {  -3.14, 0.0625, -0.537, 1.06 },
                {   4.01, 0.0421, -0.884, 1.34 },
                {   2.83, 0.0164, -0.375, 1.60 },
                {  -1.09, 0.0118, -0.143, 0.83 },
                {   2.59, 0.0291, -0.264, 0.78 },
                {   1.31, 0.0136,  0.581, 0.65 },
                {   4.08, 0.0142,  0.321, 0.65 },
                {   3.77, 0.0084,  0.219, 0.97 },
                {   0.86, 0.0878, -2.171, 2.75 },
                {  -2.02, 0.0253, -0.548, 0.68 },
                {  -3.00, 0.0204, -0.658, 1.18 },
                {  -7.95, 0.0095, -0.283, 1.33 },
                {   3.54, 0.0592, -0.752, 1.35 },
                {   3.91, 0.0872, -1.002, 1.14 },
                {  -1.11, 0.0040, -0.305, 0.91 },
                { -11.04, 0.0265, -0.409, 0.93 },
                {   3.27, 0.0689, -1.163, 1.56 },
                {  -6.89, 0.0663, -0.678, 2.17 },
                {   1.12, 0.0448, -0.321, 1.40 },
                {   3.26, 0.0076,  0.161, 0.80 },
                {   2.00, 0.0056,  0.334, 0.63 },
                {  -0.25, 0.0255,  0.751, 0.66 },
                {   0.51, 0.0069,  0.126, 0.53 },
                {  -2.28, 0.0138,  0.373, 0.92 },
                {  -2.61, 0.0264,  0.446, 0.88 },
                {  -9.24, 0.0299, -0.309, 0.79 },
            };
            #endregion

            // A ColorScheme is not needed because all points have a valid Brush
            cScatterData i_Data = new cScatterData(eScatterMode.Shapes, null);

            for (int P = 0; P < d_Values.GetLength(0); P++)
            {
                double d_Value = d_Values[P, 0];
                int s32_Radius = (int)Math.Abs(d_Value) + 1;

                double X = d_Values[P,1];
                double Y = d_Values[P,2];
                double Z = d_Values[P,3];

                eScatterShape e_Shape = (d_Value < 0) ? eScatterShape.Square : eScatterShape.Triangle;
                Brush         i_Brush = (d_Value < 0) ? Brushes.Red          : Brushes.Lime;

                String s_Tooltip = "Value = " + Graph3D.FormatDouble(d_Value);
                i_Data.AddShape(X, Y, Z, e_Shape, s32_Radius, i_Brush, s_Tooltip);
            }

            cMessgData i_Mesg1 = new cMessgData("Negative Values (size of square represents value)",   10, -10, Color.Red);
            cMessgData i_Mesg2 = new cMessgData("Positive Values (size of triangle represents value)", 10, -27, Color.Lime);

            graph3D.BeginUpdate();
            graph3D.AddRenderData (i_Data);
            graph3D.AddMessageData(i_Mesg1);
            graph3D.AddMessageData(i_Mesg2);
            graph3D.EndUpdate(eNormalize.Separate);
        }

        /// <summary>
        /// This demonstrates how to set X, Y, Z scatterplot points in form of a heart
        /// </summary>
        private void LoadValentine()
        {
            // red line of 5 pixels width 
            cColorScheme i_Scheme = new cColorScheme(5, Color.Red);
            cScatterData i_Data   = new cScatterData(eScatterMode.Lines, i_Scheme);

            // Upper (round) part of heart
            double X = 0.0;
            double Z = 0.0;
            for (double P = 0.0; P <= Math.PI * 1.32; P += 0.025)
            {
                X = Math.Cos(P) * 1.5 - 1.5;
                Z = Math.Sin(P) * 3.0 + 6.0;
                i_Data.AddLine( X, -X, Z, true,  null); // right side
                i_Data.AddLine(-X,  X, Z, false, null); // left  side
            }

            // Lower (linear) part of heart
            double d_X = X / 70;
            double d_Z = Z / 70;
            while (Z >= 0.0)
            {
                i_Data.AddLine( X, -X, Z, true,  null); // right side
                i_Data.AddLine(-X,  X, Z, false, null); // left  side
                X -= d_X;
                Z -= d_Z;
            }

            cMessgData i_Mesg = new cMessgData("Happy Valentine's day, Sweetheart!", -10, -10, Color.Red);

            graph3D.BeginUpdate();
            graph3D.AddRenderData (i_Data);
            graph3D.AddMessageData(i_Mesg);
            graph3D.EndUpdate(eNormalize.MaintainXYZ);
        }

        private void LoadPyramid()
        {
            Color[]   c_Colors = ColorSchema.GetSchema((eSchema)comboColors.SelectedIndex);
            cLineData i_Data   = new cLineData(new cColorScheme(4, c_Colors));

            cPoint3D i_Center  = new cPoint3D(25, 25, 40);
            cPoint3D i_Corner1 = new cPoint3D(25,  5,  5);
            cPoint3D i_Corner2 = new cPoint3D( 5, 25,  5);
            cPoint3D i_Corner3 = new cPoint3D(25, 45,  5);
            cPoint3D i_Corner4 = new cPoint3D(45, 25,  5);

            // Add the 4 vertical lines which are rendered as 50 parts with different colors
            const int PARTS = 50;
            i_Data.AddMultiColorLine(i_Center, i_Corner1, PARTS);
            i_Data.AddMultiColorLine(i_Center, i_Corner2, PARTS);
            i_Data.AddMultiColorLine(i_Center, i_Corner3, PARTS);
            i_Data.AddMultiColorLine(i_Center, i_Corner4, PARTS);

            // Add the 4 base lines with solid color
            i_Data.AddSolidLine(i_Corner1, i_Corner2, null);
            i_Data.AddSolidLine(i_Corner2, i_Corner3, null);
            i_Data.AddSolidLine(i_Corner3, i_Corner4, null);
            i_Data.AddSolidLine(i_Corner4, i_Corner1, null);

            graph3D.BeginUpdate();
            graph3D.AddRenderData(i_Data);
            graph3D.EndUpdate(eNormalize.MaintainXYZ);
        }

        private void LoadSphere(eSurfaceMode e_Mode, bool b_Closed)
        {
            const int LONGI  = 50; // count of 3D points along the longitude (360 degree)
            const int LATI   = 25; // count of 3D points along the latitude  (180 degree)
            const int RADIUS = 20;

            // ------ Calculate 3D points ------ 

            cPoint3D[,] i_Points = new cPoint3D[LONGI, LATI];

            for (int Long=0; Long<LONGI; Long++)
            {
                double d_Theta = 2 * Math.PI / LONGI * Long;

                for (int Lati=0; Lati<LATI; Lati++)
                {
                    double d_Phi = Math.PI / LATI * Lati;

                    // Cartesian coordinates
                    double X = RADIUS * Math.Sin(d_Phi) * Math.Cos(d_Theta);
                    double Y = RADIUS * Math.Sin(d_Phi) * Math.Sin(d_Theta);
                    double Z = RADIUS * Math.Cos(d_Phi);

                    i_Points[Long, Lati] = new cPoint3D(X, Y, Z);
                }
            }

            // ------ Create rectangular 3D polygons ------ 

            Color[]    c_Colors = ColorSchema.GetSchema((eSchema)comboColors.SelectedIndex);
            cPolygonData i_Data = new cPolygonData(e_Mode, new cColorScheme(2, c_Colors));

            // Omit top and bottom where ploygons become very narrow.
            // In case of the open sphere, omit 4 rows of rectangles so the interior becomes visible.
            int s32_Omit = b_Closed ? 1 : 4;

            for (int Long=0; Long<LONGI; Long++) // 0 .... 360 degree
            {
                // overflow to zero if maximum exceeded
                int NextLong = (Long + 1) % LONGI; 

                for (int Lati=s32_Omit; Lati<LATI-s32_Omit; Lati++) 
                {
                    int NextLati = Lati + 1;

                    List<cPoint3D> i_Poly = new List<cPoint3D>();
                    i_Poly.Add(i_Points[Long,     Lati]);
                    i_Poly.Add(i_Points[NextLong, Lati]);
                    i_Poly.Add(i_Points[NextLong, NextLati]);
                    i_Poly.Add(i_Points[Long,     NextLati]);
                    i_Data.AddPolygon(i_Poly);
                }
            }

            // ----- Close the top and bottom with a round polygon ------

            if (e_Mode == eSurfaceMode.Fill && b_Closed)
            {
                List<cPoint3D> i_Top    = new List<cPoint3D>();
                List<cPoint3D> i_Bottom = new List<cPoint3D>();
                for (int Long=0; Long<LONGI; Long++) // 0 .... 360 degree
                {
                    i_Top   .Add(i_Points[Long, 1]);
                    i_Bottom.Add(i_Points[Long, LATI-1]);
                }
                i_Data.AddPolygon(i_Top);
                i_Data.AddPolygon(i_Bottom);
            }

            graph3D.BeginUpdate();
            graph3D.AddRenderData(i_Data);
            graph3D.EndUpdate(eNormalize.MaintainXYZ);
        }
    }
}
